#!/usr/bin/env python
"""
Map2KO

Genome library

Handle a KO search on a local machine (Blast-BBH) or online (KAAS)
"""
from Bio import SeqIO
from ductape.common.commonmultiprocess import CommonMultiProcess
from ductape.common.utils import slice_it
from ductape.genome.blast import Blaster, RunBBH
import Queue
import logging
import os
import webbrowser

__author__ = "Marco Galardini"

################################################################################
# Log setup

logger = logging.getLogger('ductape.map2KO')

################################################################################
# Classes

class KOBBH(object):
    pass

class LocalSearch(CommonMultiProcess):
    '''
    Class localSearch
    '''
    _statusDesc = {0:'Not started',
               1:'Making room', 
               2:'Creating Blast DB',
               3:'Running Blast',
               4:'Running Blast on short proteins',
               5:'Parsing Blast',
               6:'Running BBH',
               7:'Cleaning up'}
    
    _substatuses = [3,4,6]
    
    def __init__(self,query,target,
                 ncpus=1,evalue=1e-50,
                 buildDB=True,bbh=True,recover=False,queue=Queue.Queue()):
        CommonMultiProcess.__init__(self,ncpus,queue)
        # Blast
        self.query = query
        if buildDB:
            self.target = target
            self.db = None
        else:
            self.target = None
            self.db = target
        self.out = []
        self.evalue = float(evalue)
        self.bbh = bool(bbh)
        self.recover = recover
        self.ncpus = int(ncpus)
        self._kohits = []
        self.results = {}
        self._keggroom = None
        self._blast = Blaster()
        
    def makeRoom(self,location=''):
        '''
        Creates a tmp directory in the desired location
        '''
        try:
            path = os.path.abspath(location)
            path = os.path.join(path, 'tmp')
            try:os.mkdir(path)
            except:pass
            path = os.path.join(path, 'blast')
            self._room = path
            os.mkdir(path)
        except:
            logger.debug('Temporary directory creation failed! %s'
                          %path)
        
        # KEGG database path
        try:
            path = os.path.abspath(location)
            path = os.path.join(path, 'tmp')
            try:os.mkdir(path)
            except:pass
            path = os.path.join(path, 'keggdb')
            self._keggroom = path
            os.mkdir(path)
        except:
            logger.debug('Temporary directory creation failed! %s'
                          %path)
        
    def createDB(self):
        self.db = os.path.join(self._keggroom,'KEGGdb') 
        return self._blast.createDB(self.target, 'prot', self.db)
    
    def runBlast(self,short=False):
        lS = []
        for s in SeqIO.parse(open(self.query),'fasta'):
            if short and len(s) <= 30:
                lS.append(s)
            elif not short:lS.append(s)
        self._maxsubstatus = len(lS)
        for seqs in slice_it(lS,10):
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return False
            
            if len(seqs) == 0:
                continue
            self._substatus += len(seqs)
            self.updateStatus(sub=True)
            if short:
                query = os.path.join(self._room,
                         'KEGGshort_%d.faa'%self._substatus)
                out = os.path.join(self._room,
                           'KEGGshort_%d.xml'%self._substatus)
            else:
                query = os.path.join(self._room,
                         'KEGG_%d.faa'%self._substatus)
                out = os.path.join(self._room,'KEGG_%d.xml'%self._substatus)
            self.out.append(out)
            # If recovery, skip the unnecessary scans
            if ( self.recover and os.path.exists(query) and
                 os.path.exists(out)):
                # Last test: can it be parsed?
                try:
                    self._blast.parseBlast(out)
                    for hits in self._blast.getHits(self.evalue):
                        pass
                    logger.debug('Skipping slice %s because has already been done'
                                %query)
                    continue
                except:
                    pass
            oseqs = SeqIO.write(seqs,open(query,'w'),'fasta')
            if oseqs != len(seqs):
                logger.warning('Query splitting error! Expected %d, '+
                                'Printed %d'%(len(seqs),oseqs))
            if short:
                res = self._blast.runBlast(query, self.db, out,
                                 evalue = self.evalue,
                                 ncpus = self.ncpus, task='blastp-short')
            else:
                res = self._blast.runBlast(query, self.db, out,
                                 evalue = self.evalue,
                                 ncpus = self.ncpus)
            if not res:
                return False
        return True
    
    def parseBlast(self):
        for out in self.out:
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return False
        
            self._blast.parseBlast(out)
            # Catch the exceptions if the XML is dirty
            try:
                for hits in self._blast.getHits(self.evalue):
                    for hit in hits:
                        if hit.getKO():
                            self._kohits.append(hit)
            except:
                logger.error('Blast results corrupted for file %s'%out)
                return False
        return True
    
    def runBBH(self):
        # Create a DB of the source genome
        sourceDB = os.path.join(self._room,'SOURCEdb') 
        if not self._blast.createDB(self.query, 'prot', sourceDB):
            logger.error('Could not create source DB %s'%sourceDB)
            return False
        self._maxsubstatus = len(self._kohits)
        
        self.initiateParallel()
        
        for hit in self._kohits:
            uniqueid = self.getUniqueID()
            
            if hit.query_len > 30:
                short = False
            else:
                short = True
            
            # Multi process
            obj = RunBBH('Map2KO',hit.query_id,sourceDB,
                    self.db,None,
                    self.evalue,'BLOSUM62',short,uniqueid,
                    kegg = True, ko_entry = hit.hit, ko_id = hit.getKO())
            self._paralleltasks.put(obj)
            
        # Poison pill to stop the workers
        self.addPoison()
        
        while True:
            while not self._parallelresults.empty():
                if self.killed:
                    logger.debug('Exiting for a kill signal')
                    return False
                
                self._substatus += 1
                self.updateStatus(sub=True)
                
                result = self._parallelresults.get()
                
                if not result[2]:
                    logger.error('An error occurred for BBH!')
                    return False
                if result[1] not in self.results and result[0]:
                    self.results[result[1]] = []
                if result[0] and result[0] not in self.results[result[1]]:
                    self.results[result[1]].append(result[0])
                    
            if self.isTerminated():
                break
            
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return False
            
            self.sleeper.sleep(0.1)
            
        # Get the last messages
        while not self._parallelresults.empty():
            if self.killed:
                logger.debug('Exiting for a kill signal')
                return False
            
            self._substatus += 1
            self.updateStatus(sub=True)
            
            result = self._parallelresults.get()
            
            if not result[2]:
                logger.error('An error occurred for BBH!')
                return False
            if result[1] not in self.results and result[0]:
                self.results[result[1]] = []
            if result[0] and result[0] not in self.results[result[1]]:
                self.results[result[1]].append(result[0])
            
        self.killParallel()            

        return True
            
    def run(self):
        self.updateStatus()
        self.makeRoom()

        if self.killed:
            return

        if not self.db:
            self.updateStatus()
            if not self.createDB():
                self.sendFailure('CreateDB failure')
                return
        else:
            self.updateStatus(send=False)
            
        if self.killed:
            return

        self.updateStatus()
        if not self.runBlast():
            self.sendFailure('RunBlast failure')
            return
        self.resetSubStatus()
        
        if self.killed:
            return
        
        self.updateStatus()
        if not self.runBlast(True):
            self.sendFailure('RunBlast (short) failure')
            return
        self.resetSubStatus()
        
        if self.killed:
            return
        
        self.updateStatus()
        if not self.parseBlast():
            self.sendFailure('ParseBlast failure')
            return
        if len(self._kohits) == 0:
            logger.warning('No protein in %s with homology to KO!'%self.query)
            self.sendFailure('No protein in %s with homology to KO!'%self.query)
            self.cleanUp()
            return
        
        if self.killed:
            return
        
        if self.bbh:
            self.updateStatus()
            if not self.runBBH():
                self.sendFailure('BBH failure')
        else:
            for hit in self._kohits:
                if hit.query_id not in self.results:
                    self.results[hit.query_id] = []
                ko = hit.getKO()
                if ko not in self.results[hit.query_id]:
                    self.results[hit.query_id].append(ko)
            self.updateStatus(send=False)   
        self.resetSubStatus()
        
        if self.killed:
            return
        
        try:
            del self.results[None]
        except:pass
        
        # Only ONE KO for each protein
        for k in self.results:
            self.results[k] = self.results[k][0]
        if len(self.results) == 0:
            logger.warning('No protein in %s with BBH to KO!'%self.query)
            self.sendFailure('No protein in %s with BBH to KO!'%self.query)
            self.cleanUp()
            return
        
        if self.killed:
            return
        
        self.updateStatus()
        self.cleanUp()

class OnlineSearch(object):
    '''
    Online KO search using KAAS
    '''
    def __init__(self,bbh=True):
        self.bbh = bbh
        self.results = {}
        self.url = 'http://www.genome.ad.jp/tools/kaas/'
        
    def getExplanation(self):
        msg = ' '.join(['The only way to access to KAAS annotation service is',
                'your browser.\nGo to '+self.url+
                ', where you can add your fasta file(s).',
                '\nThe analysis will take less than one hour.'])
        if self.bbh:
            msg = '\n'.join([msg, 'Select the BBH option and use the appropriate'+
                            ' organisms list.'])
        else:
            msg = '\n'.join([msg, 'Select the SBH option and use the appropriate'+
                            ' organisms list.'])
        msg = '\n'.join([msg, 'When the analysis is finished save the KO list file.'])
        return msg
    
    def openBrowser(self):
        webbrowser.open(self.url, new=2)
        
    def parseKAAS(self,filein):
        for l in open(filein):
            if l.startswith('#'):continue
            s=l.strip().split('\t')
            if len(s) == 1:continue
            self.results[s[0]] = s[1]


#!/usr/bin/env python
"""
Net

Kegg Library

Networks made using Kegg data
"""
import logging
import networkx as nx
import numpy as np
# Nodes color handling
import matplotlib.colors as pltcls
import matplotlib.pyplot as plt

__author__ = "Marco Galardini"

################################################################################
# Log setup

logger = logging.getLogger('ductape.net')

################################################################################
# Classes

class Compound(object):
    '''
    Kegg compound, enconded as a node
    '''
    def __init__(self, co_id, name, weight=None, vmax=9):
        self.co_id = co_id
        self.name = name
        
        self.vmax = vmax

        if weight is not None:
            self.weight = weight
            
    def getColor(self):
        '''
        Transform the node weight into an hex color
        '''
        phenomeNorm = pltcls.Normalize(vmin=0, vmax=self.vmax)
        if self.weight:
            return pltcls.rgb2hex( 
                                plt.get_cmap('RdYlGn')(phenomeNorm(self.weight))
                                )
        
        return '#808080'
                    
class Reaction(object):
    '''
    Kegg reaction, enconded as an edge
    '''
    def __init__(self, re_id, co1, co2, name, weight=None):
        self.re_id = re_id
        self.co1 = co1
        self.co2 = co2
        self.name = name
        
        if weight:
            self.weight = weight

class MetabolicNet(object):
    '''
    A metabolic network as a networkX graph
    
    Nodes: kegg compounds (w co_id, name; w or w/o weight)
    Edges: kegg reactions (w co1, co2, re_id, name; w or w/o weight)
    
    nodes weight indicate the activity index
    edges weight indicate the copy number 
    '''
    def __init__(self, nodes=None, edges=None, name='MetNet'):
        self.name = name
        
        self.net = nx.Graph()
        
        if nodes:
            for n in nodes:
                self.net.add_node(n.co_id, name=n.name)
                if hasattr(n, 'weight'):
                    self.net.node[n.co_id]['weight'] = n.weight
                    self.net.node[n.co_id]['graphics'] = {'fill': n.getColor()}
        
        if edges:
            for e in edges:
                self.net.add_edge(e.co1, e.co2, re_id=e.re_id, name=e.name)
                if hasattr(e, 'weight'):
                    self.net.edge[e.co1][e.co2]['weight'] = e.weight
                    
    def hasNodesWeight(self):
        '''
        At least one node has weight?
        '''
        for co in self.net.nodes():
            try:
                self.net.node[co]['weight']
                return True
            except:pass
            
        return False
    
    def hasEdgesWeight(self):
        '''
        At least one edge has weight?
        '''
        for e in self.net.edges():
            if 'weight' in self.net.edge[e[0]][e[1]]:
                return True

        return False
    
    def removeSingletons(self):
        '''
        Remove nodes with degree 0
        '''
        to_remove = filter(lambda x: self.net.degree()[x] == 0,
                           self.net.nodes())
        self.net.remove_nodes_from(to_remove)
            
    def setNet(self, net):
        '''
        Use an external networkx graph
        '''
        self.net = net
                    
    def addNodes(self, nodes):
        '''
        Takes a compounds iterable and adds (or updates) the nodes
        w co_id, name; w or w/o weight attributes
        nodes weight indicate the activity index
        '''
        for n in nodes:
            self.net.add_node(n.co_id, name=n.name)
            if hasattr(n, 'weight'):
                self.net.node[n.co_id]['weight'] = n.weight
                self.net.node[n.co_id]['graphics'] = {'fill': n.getColor()}
                
    def __len__(self):
        '''
        Returns the number of reactions
        '''
        i = 0
        for e in self.net.edges():
            if 'weight' in self.net.edge[e[0]][e[1]]:
                i += self.net.edge[e[0]][e[1]]['weight']
            else:
                i += 1
                
        return i
    
    def getDistinctReactions(self):
        '''
        Returns the distinct reaction IDs of this network
        '''
        re_id = set()
        for e in self.net.edges():
            re_id.add(self.net.edge[e[0]][e[1]]['re_id'])
        
        return re_id
    
    def mean(self):
        '''
        Get the mean compounds activity (nodes weight)  
        '''
        weights = []
        
        for co in self.net.nodes():
            try:
                weights.append(self.net.node[co]['weight'])
            except:pass
            
        if len(weights) == 0:
            return np.nan
        
        return np.array(weights).mean()
    
    def std(self):
        '''
        Get the stddev compounds activity (nodes weight)  
        '''
        weights = []
        
        for co in self.net.nodes():
            try:
                weights.append(self.net.node[co]['weight'])
            except:pass
            
        if len(weights) == 0:
            return np.nan
        
        return np.array(weights).std()

    def getComponents(self):
        return len(nx.connected_components(self.net))
    
    def getComponentsSizes(self):
        return [len(x) for x in nx.connected_components(self.net)]
    
    def getComponentsMean(self):
        return np.array([len(x)
                         for x in nx.connected_components(self.net)]).mean()
                         
    def getComponentsStd(self):
        return np.array([len(x)
                         for x in nx.connected_components(self.net)]).std()



#!/usr/bin/env python
"""
clustering

Phenome library

biolog data clustering functions
Many thanks to the scikits.learn team for the exhaustive documentation
"""
from itertools import product
from sklearn.cluster import KMeans, MeanShift, estimate_bandwidth
import numpy as np
import logging
import warnings

__author__ = "Marco Galardini"

logger = logging.getLogger('ductape.clustering')

def plotElbow(d, param_labels):
    from scipy.interpolate import interp1d
    import matplotlib.pyplot as plt
    
    logger.info('Saving elbow test plots')
    
    x_new = np.linspace(-1, max(d.keys()), 100)
    
    figidx = 1
    
    figsize = (len(d[d.keys()[0]][0])/2) + (len(d[d.keys()[0]][0])%2)  
    
    fig = plt.figure(figsize=(3.5*figsize, 8))
    fig.clf()
    for j in range(len(d[d.keys()[0]][0])):
        ax = fig.add_subplot(2, figsize, figidx)
        
        figidx += 1
        
        diffs = {}
        for i in d:
            diffs[i] = np.array([p[j] for p in d[i]]).mean()
        
        inter = interp1d(d.keys(), [diffs[i] for i in d], bounds_error=False,
                     kind='cubic')
        ax.plot(d.keys(),[diffs[i] for i in d],'o', x_new, inter(x_new),'-')
        ax.set_ylabel('Sum of squared errors')
        ax.set_xlabel('Num. clusters')
        ax.set_title(param_labels[j])
        
    fig.tight_layout()
    fig.savefig('elbow.png',dpi=300)

def plotClusters(X, labels, params=None, method='', prefix='clusters'):
    from ductape.common.utils import slice_it
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
    
    X = np.array(X)
    labels = np.array(labels)
    labels_unique = np.unique(labels)
    n_clusters_ = len(labels_unique)
    
    figidx = 1
    fig = plt.figure(1)
    fig.clf()
    for x, y in product(range(len(X[0])), repeat=2):
        ax = fig.add_subplot(len(X[0]),len(X[0]),figidx)
        
        if figidx%len(X[0]) == 1:
            if not params:
                ax.set_ylabel(x, rotation='horizontal')
            else:
                ax.set_ylabel(params[x], rotation='horizontal')
        if abs((len(X[0])*len(X[0])) % figidx - len(X[0])) <= len(X[0]):
            if not params:
                ax.set_xlabel(y)
            else:
                ax.set_xlabel(params[y])
        
        figidx += 1

        color = dict()
        j = 0
        for i in slice_it(range(255), cols=n_clusters_):
            color[j] = cm.RdYlGn(i[0])
            j += 1

        for k in range(n_clusters_):
            my_members = labels == k
            ax.plot(X[my_members, y], X[my_members, x], '.', color=color[k])

    fig.subplots_adjust(wspace=0, hspace=0)
    for ax in fig.axes:
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])
    fig.suptitle('Clusters (%s, %s): %d' % (prefix, method, n_clusters_))
    fig.savefig('%s_%s.png'%(prefix,method),dpi=300)

def mean(X, save_fig=False, params_labels=None, prefix='clusters'):
    '''
    Compute clustering with MeanShift
    '''
    logger.debug('Calculating MeanShift clusters using %d parameters'%len(X[0]))
    
    X = np.array( X )
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        bandwidth = estimate_bandwidth(X, quantile=0.2)
    
        ms = MeanShift(bandwidth=bandwidth, bin_seeding=True)
        ms.fit(X)
        
    labels = ms.labels_
    
    if save_fig:
        plotClusters(X, ms, method='mean', prefix=prefix,
                     params=params_labels)
    
    labels_unique = np.unique(labels)
    n_clusters_ = len(labels_unique)
    
    logger.debug('Found %d clusters with MeanShift algorithm'%n_clusters_)
    
    return labels

def _kmeans(X, n_clusters):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        
        try:
            k_means = KMeans(init='random', k=n_clusters, n_init=100,
                             max_iter=1000)
        except:
            k_means = KMeans(init='random', n_clusters=n_clusters, n_init=100,
                             max_iter=1000)
        k_means.fit(X)
        
    return k_means

def getSseKmeans(k_means, X):
    '''
    Returns a list of sums of squared errors
    '''
    import math
    
    sse=[]
    for j in range(len(k_means.labels_)):
        c = k_means.labels_[j]
        ref = k_means.cluster_centers_[c]
        dist = []
        for k in range(len(ref)):
            dist.append(math.pow(X[j][k] - ref[k], 2))
        sse.append(dist)
        
    return sse

def kmeans(X, n_clusters=10, save_fig=False, params_labels=None, prefix='clusters'):
    '''
    Compute clustering with KMeans
    '''
    logger.debug('Calculating KMean clusters using %d parameters'%len(X[0]))
    
    X = np.array( X )
    
    k_means = _kmeans(X, n_clusters)
    
    labels = k_means.labels_
    
    if save_fig:
        plotClusters(X, k_means, method='kmeans', prefix=prefix,
                     params=params_labels)
    
    labels_unique = np.unique(labels)
    n_clusters_ = len(labels_unique)
    
    logger.debug('Found %d clusters with KMeans algorithm'%n_clusters_)
    
    return labels
    